"""Core recipes for Gaussian."""

from __future__ import annotations

from typing import TYPE_CHECKING

import psutil

from quacc import Remove, job
from quacc.recipes.gaussian._base import _LABEL, run_and_summarize

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc.types import Filenames, RunSchema, SourceDirectory


@job
def static_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    xc: str = "wb97xd",
    basis: str = "def2tzvp",
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
    **calc_kwargs,
) -> RunSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.
    **calc_kwargs
        Custom kwargs for the Gaussian calculator. Set a value to
        `quacc.Remove` to remove a pre-existing key entirely. For a list of available
        keys, refer to the [ase.calculators.gaussian.Gaussian][] calculator.

    Returns
    -------
    RunSchema
        Dictionary of results
    """
    calc_defaults = {
        # General settings
        "mem": "16GB",
        "chk": f"{_LABEL}.chk",
        "nprocshared": psutil.cpu_count(logical=False),
        "xc": xc,
        "basis": basis,
        "charge": charge,
        "mult": spin_multiplicity,
        "scf": ["maxcycle=250", "xqc"],
        "integral": "ultrafine",
        "pop": "CM5",
        # Job-specific settings
        "force": "",
        "gfinput": "",
        "ioplist": ["6/7=3", "2/9=2000"],  # https://gitlab.com/ase/ase/-/issues/660
    }
    return run_and_summarize(
        atoms,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "Gaussian Static"} | (additional_fields or {}),
        copy_files=copy_files,
    )


@job
def relax_job(
    atoms: Atoms,
    charge: int,
    spin_multiplicity: int,
    xc: str = "wb97xd",
    basis: str = "def2tzvp",
    freq: bool = False,
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
    **calc_kwargs,
) -> RunSchema:
    """
    Carry out a geometry optimization.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set
    freq
        If a frequency calculation should be carried out.
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results
    **calc_kwargs
        Custom kwargs for the Gaussian calculator. Set a value to
        `quacc.Remove` to remove a pre-existing key entirely. For a list of available
        keys, refer to the [ase.calculators.gaussian.Gaussian][] calculator.

    Returns
    -------
    RunSchema
        Dictionary of results
    """
    calc_defaults = {
        # General settings
        "mem": "16GB",
        "chk": f"{_LABEL}.chk",
        "nprocshared": psutil.cpu_count(logical=False),
        "xc": xc,
        "basis": basis,
        "charge": charge,
        "mult": spin_multiplicity,
        "scf": ["maxcycle=250", "xqc"],
        "integral": "ultrafine",
        "pop": "CM5",
        # Job-specific settings
        "opt": "",
        "nosymmetry": "",
        "ioplist": ["2/9=2000"],  # https://gitlab.com/ase/ase/-/issues/660
        "freq": "" if freq else Remove,
    }

    return run_and_summarize(
        atoms,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "Gaussian Relax"} | (additional_fields or {}),
        copy_files=copy_files,
    )



"""Schemas for storing metadata about Atoms objects."""

from __future__ import annotations

from typing import TYPE_CHECKING

from emmet.core.structure import MoleculeMetadata, StructureMetadata
from pymatgen.io.ase import AseAtomsAdaptor

from quacc.atoms.core import copy_atoms

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc.types import AtomsSchema


def atoms_to_metadata(
    atoms: Atoms, additional_fields: dict[str, Any] | None = None
) -> AtomsSchema:
    """
    Convert an ASE Atoms object to a dict suitable for storage in MongoDB.

    Parameters
    ----------
    atoms
        ASE Atoms object to store in {"atoms": atoms}
    additional_fields
        Additional fields to add to the document.

    Returns
    -------
    AtomsSchema
        Dict of metadata about the Atoms object.
    """
    additional_fields = additional_fields or {}
    atoms = copy_atoms(atoms)
    results = {}
    atoms.calc = None

    # Strip the dummy atoms, if present
    if "X" in atoms.get_chemical_symbols():
        del atoms[[atom.index for atom in atoms if atom.symbol == "X"]]

    # Get Pymatgen Structure/Molecule metadata
    if atoms.pbc.any():
        structure = AseAtomsAdaptor().get_structure(atoms)
        structure_metadata = StructureMetadata().from_structure(structure).model_dump()
        results["structure_metadata"] = structure_metadata
    else:
        mol = AseAtomsAdaptor().get_molecule(atoms, charge_spin_check=False)
        molecule_metadata = MoleculeMetadata().from_molecule(mol).model_dump()
        for key in ["charge", "spin_multiplicity", "nelectrons"]:
            del molecule_metadata[key]
        results["molecule_metadata"] = molecule_metadata

    # Store Atoms object
    results["atoms"] = atoms

    return results | additional_fields



"""
Meta FAIR recipes

!!! Important

    Make sure that you use the appropriate version pseudpotentials:
    - OMat: v.54
    - OMC: v.54
"""

from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING

from monty.dev import requires

from quacc import job
from quacc.calculators.vasp.params import MPtoASEConverter
from quacc.recipes.vasp._base import run_and_summarize

has_fairchem = bool(find_spec("fairchem"))
has_fairchem_omat = has_fairchem and bool(find_spec("fairchem.data.omat"))
has_atomate2 = bool(find_spec("atomate2"))

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc.types import Filenames, SourceDirectory, VaspSchema


@job
@requires(
    has_fairchem_omat,
    "fairchem-data-omat is not installed. Run `pip install quacc[fairchem]`",
)
def omat_static_job(
    atoms: Atoms,
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
    **calc_kwargs,
) -> VaspSchema:
    """
    Carry out a static calculation with OMat settings.

    Parameters
    ----------
    atoms
        Atoms object
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.
    **calc_kwargs
        Custom kwargs for the Vasp calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to [quacc.calculators.vasp.vasp.Vasp][]. All of the ASE
        Vasp calculator keyword arguments are supported.

    Returns
    -------
    VaspSchema
        Dictionary of results from [quacc.schemas.vasp.VaspSummarize.run][].
        See the type-hint for the data structure.
    """
    from fairchem.data.omat.vasp.sets import OMat24StaticSet

    calc_defaults = MPtoASEConverter(atoms=atoms).convert_input_set(OMat24StaticSet())
    calc_defaults["pp_version"] = "54"

    return run_and_summarize(
        atoms,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "OMat Static"} | (additional_fields or {}),
        copy_files=copy_files,
    )


@job
@requires(has_atomate2, "atomate2 is not installed. Run `pip install quacc[fairchem]`")
def omc_static_job(
    atoms: Atoms,
    copy_files: SourceDirectory | dict[SourceDirectory, Filenames] | None = None,
    additional_fields: dict[str, Any] | None = None,
    **calc_kwargs,
) -> VaspSchema:
    """
    Carry out a static calculation with OMC settings.

    Parameters
    ----------
    atoms
        Atoms object
    copy_files
        Files to copy (and decompress) from source to the runtime directory.
    additional_fields
        Additional fields to add to the results dictionary.
    **calc_kwargs
        Custom kwargs for the Vasp calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to [quacc.calculators.vasp.vasp.Vasp][]. All of the ASE
        Vasp calculator keyword arguments are supported.

    Returns
    -------
    VaspSchema
        Dictionary of results from [quacc.schemas.vasp.VaspSummarize.run][].
        See the type-hint for the data structure.
    """

    calc_defaults = _make_omc_inputs(atoms)
    calc_defaults["pp_version"] = "54"

    return run_and_summarize(
        atoms,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "OMC Static"} | (additional_fields or {}),
        copy_files=copy_files,
    )


def _make_omc_inputs(atoms: Atoms) -> dict:
    """
    Helper function to make a fairchem input set.

    Parameters
    ----------
    atoms
        Atoms object
    dataset
        Dataset to use. Currently only "omc" is supported.

    Returns
    -------
    dict
        Dictionary of ASE VASP calculator parameters.
    """
    from atomate2.vasp.sets.core import StaticSetGenerator

    input_generator = StaticSetGenerator(
        user_incar_settings={
            "ADDGRID": True,
            "ALGO": "Normal",
            "EDIFF": 1e-06,
            "ENCUT": 520,
            "GGA": "PE",
            "IBRION": -1,
            "ISIF": 0,
            "ISMEAR": 0,
            "ISPIN": 1,
            "IVDW": 11,
            "LREAL": False,
            "LMIXTAU": True,
            "LASPH": True,
            "LORBIT": 11,
            "LWAVE": False,
            "LAECHG": False,
            "LVTOT": False,
            "NCORE": 100,
            "NELM": 200,
            "NELMDL": -10,
            "NSW": 0,
            "PREC": "Normal",
            "SIGMA": 0.1,
        },
        user_potcar_functional="PBE_54_W_HASH",
        auto_kspacing=True,
    )
    return MPtoASEConverter(atoms=atoms).convert_input_generator(input_generator)



"""Common elastic constants workflows."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ase import units
from ase.stress import voigt_6_to_full_3x3_stress
from emmet.core.elasticity import ElasticityDoc
from pymatgen.analysis.elasticity.stress import Stress
from pymatgen.io.ase import AseAtomsAdaptor

from quacc import flow, job, subflow
from quacc.atoms.deformation import make_deformations_from_bulk

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms
    from pymatgen.analysis.elasticity.strain import DeformedStructureSet

    from quacc import Job
    from quacc.types import ElasticSchema, RunSchema


@flow
def elastic_tensor_flow(
    atoms: Atoms,
    relax_job: Job,
    static_job: Job,
    pre_relax: bool = True,
    run_static: bool = False,
    deform_kwargs: dict[str, Any] | None = None,
) -> ElasticSchema:
    """
    Common workflow for calculating elastic tensors.

    Parameters
    ----------
    atoms
        Atoms object
    relax_job
        The relaxation function.
    static_job
        The static function
    pre_relax
        Whether to run a relaxation on the bulk structure before deformation (true) or run a static
        calculation (false)
    run_static
        Whether to run static calculations after any relaxations on the undeformed or deformed structures
    deform_kwargs
        Additional keyword arguments to pass to [quacc.atoms.deformation.make_deformations_from_bulk][]

    Returns
    -------
    ElasticSchema
        See the return type-hint for the data structure.
    """
    if pre_relax:
        undeformed_result = relax_job(atoms, relax_cell=True)
        if run_static:
            undeformed_result = static_job(undeformed_result["atoms"])
    else:
        undeformed_result = static_job(atoms)

    return _elastic_tensor_subflow(
        undeformed_result=undeformed_result,
        relax_job=relax_job,
        static_job=static_job if run_static else None,
        deform_kwargs=deform_kwargs,
    )


@subflow
def _elastic_tensor_subflow(
    undeformed_result: RunSchema,
    relax_job: Job,
    static_job: Job | None = None,
    deform_kwargs: dict[str, Any] | None = None,
) -> ElasticSchema:
    """
    Workflow consisting of:

    1. Deformed structures generation

    2. Deformed structures relaxations

    3. Deformed structures statics (optional)

    4. Elastic tensor calculation

    Parameters
    ----------
    undeformed_result
        Result of a static or optimization calculation
    relax_job
        The relaxation function.
    static_job
        The static function
    deform_kwargs
        Additional keyword arguments to pass to
        [quacc.atoms.deformation.make_deformations_from_bulk][]

    Returns
    -------
    list[dict]
        List of schemas.
    """
    deform_kwargs = deform_kwargs or {}

    deformed_structure_set = make_deformations_from_bulk(
        undeformed_result["atoms"], **deform_kwargs
    )

    results = []
    for deformed in deformed_structure_set:
        result = relax_job(deformed.to_ase_atoms(), relax_cell=False)

        if static_job is not None:
            result = static_job(result["atoms"])

        results.append(result)

    elasticity_doc = _deformations_to_elastic_tensor(
        undeformed_result=undeformed_result,
        deformed_structure_set=deformed_structure_set,
        results=results,
    )

    return {
        "deformed_structure_set": deformed_structure_set,
        "deformed_results": results,
        "undeformed_result": undeformed_result,
        "elasticity_doc": elasticity_doc,
    }


@job
def _deformations_to_elastic_tensor(
    undeformed_result: RunSchema,
    deformed_structure_set: DeformedStructureSet,
    results: list[RunSchema],
) -> ElasticityDoc:
    """
    Function to fit a DeformedStructureSet and result documents to an elastic tensor

    Parameters
    ----------
    undeformed_result
        Single point or relaxation of the undeformed result (for the relaxed stress, if it's not quite 0)
    deformed_structure_set
        The pymatgen DeformedStructureSet with information on the strains of each structure
    results
        A list of results (one per deformed structure) corresponding to single points or relaxations
        on the deformed structures.

    Returns
    -------
    ElasticityDoc
        An emmet (i.e. MaterialsProject) ElasticityDoc with information about the final elastic tensor fit
    """
    structure = AseAtomsAdaptor.get_structure(undeformed_result["atoms"])
    equilibrium_stress = Stress(
        (
            voigt_6_to_full_3x3_stress(undeformed_result["results"]["stress"])
            if len(undeformed_result["results"]["stress"]) == 6
            else undeformed_result["results"]["stress"]
        )
        / units.GPa
    )
    stresses = [
        Stress(
            (
                voigt_6_to_full_3x3_stress(relax_result["results"]["stress"])
                if len(relax_result["results"]["stress"]) == 6
                else relax_result["results"]["stress"]
            )
            / units.GPa
        )
        for relax_result in results
    ]
    return ElasticityDoc.from_deformations_and_stresses(
        structure,
        deformations=deformed_structure_set.deformations,
        equilibrium_stress=equilibrium_stress,
        stresses=stresses,
    )



"""Schemas for storing ASE-based data."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from ase.io import read
from ase.vibrations.data import VibrationsData
from emmet.core.symmetry import PointGroupData
from pymatgen.io.ase import AseAtomsAdaptor

from quacc import QuaccDefault, __version__, get_settings
from quacc.atoms.core import get_final_atoms_from_dynamics
from quacc.schemas.atoms import atoms_to_metadata
from quacc.schemas.prep import prep_next_run
from quacc.schemas.thermo import ThermoSummarize
from quacc.utils.dicts import clean_dict, recursive_dict_merge
from quacc.utils.files import get_uri

if TYPE_CHECKING:
    from typing import Any, Literal

    from ase.atoms import Atoms
    from ase.io.trajectory import TrajectoryWriter
    from ase.md.md import MolecularDynamics
    from ase.optimize.optimize import Optimizer
    from ase.vibrations import Vibrations

    from quacc.types import (
        DefaultSetting,
        DynSchema,
        OptSchema,
        RunSchema,
        VibSchema,
        VibThermoSchema,
    )


class Summarize:
    """
    Get tabulated results from an Atoms object and calculator and store them in a
    database-friendly format. This is meant to be compatible with all calculator types.
    """

    def __init__(
        self,
        directory: str | Path | None = None,
        move_magmoms: bool = False,
        additional_fields: dict[str, Any] | None = None,
    ) -> None:
        """
        Initialize the Summarize object.

        Parameters
        ----------
        directory
            Path to the directory where the calculation was run and results were stored.
        move_magmoms
            Whether to move the final magmoms of the original Atoms object to the
            initial magmoms of the returned Atoms object, if relevant.
        additional_fields
            Additional fields to add to the task document.

        Returns
        -------
        None
        """
        self.directory = directory
        self.move_magmoms = move_magmoms
        self.additional_fields = additional_fields or {}
        self._settings = get_settings()

    def run(self, final_atoms: Atoms, input_atoms: Atoms) -> RunSchema:
        """
        Get tabulated results from a standard ASE run.

        Parameters
        ----------
        final_atoms
            ASE Atoms following a calculation. A calculator must be attached.
        input_atoms
            Input ASE Atoms object to store.

        Returns
        -------
        RunSchema
            Dictionary representation of the task document
        """

        # Check and set up variables
        if not final_atoms.calc:
            msg = "ASE Atoms object has no attached calculator."
            raise ValueError(msg)
        if not final_atoms.calc.results:
            msg = "ASE Atoms object's calculator has no results."
            raise ValueError(msg)

        directory = self.directory or final_atoms.calc.directory

        # Generate input atoms metadata
        input_atoms_metadata = atoms_to_metadata(input_atoms) if input_atoms else {}

        # Generate the base of the task document
        inputs = {
            "parameters": final_atoms.calc.parameters,
            "nid": get_uri(directory).split(":")[0],
            "dir_name": directory,
            "input_atoms": input_atoms_metadata,
            "quacc_version": __version__,
        }
        results = {"results": final_atoms.calc.results}

        # Prepare atoms for the next run
        atoms_to_store = prep_next_run(final_atoms, move_magmoms=self.move_magmoms)

        # Generate final atoms metadata
        final_atoms_metadata = atoms_to_metadata(atoms_to_store) if final_atoms else {}

        # Create a dictionary of the inputs/outputs
        unsorted_task_doc = (
            final_atoms_metadata | inputs | results | self.additional_fields
        )

        return clean_dict(unsorted_task_doc)

    def opt(
        self,
        dyn: Optimizer,
        trajectory: list[Atoms] | None = None,
        check_convergence: bool | DefaultSetting = QuaccDefault,
    ) -> OptSchema:
        """
        Get tabulated results from an ASE optimization.

        Parameters
        ----------
        dyn
            ASE Optimizer object.
        trajectory
            ASE Trajectory object or list[Atoms] from reading a trajectory file.
        check_convergence
            Whether to check the convergence of the calculation. Defaults to True in
            settings.

        Returns
        -------
        OptSchema
            Dictionary representation of the task document
        """

        # Check and set up variables
        check_convergence = (
            self._settings.CHECK_CONVERGENCE
            if check_convergence == QuaccDefault
            else check_convergence
        )

        # Get trajectory
        atoms_trajectory = (
            trajectory or read(dyn.trajectory, index=":")
            if isinstance(dyn.trajectory, (str, Path))
            else read(dyn.trajectory.filename, index=":")
        )
        trajectory_results = [atoms.calc.results for atoms in atoms_trajectory]

        initial_atoms = atoms_trajectory[0]
        final_atoms = get_final_atoms_from_dynamics(dyn)
        directory = self.directory or final_atoms.calc.directory

        # Check convergence
        gradient = (
            dyn.optimizable.get_gradient() if hasattr(dyn, "optimizable") else None
        )
        is_converged = dyn.converged(gradient)
        if check_convergence and not is_converged:
            msg = f"Optimization did not converge. Refer to {directory}"
            raise RuntimeError(msg)

        # Base task doc
        base_task_doc = self.run(final_atoms, initial_atoms)

        # Clean up the opt parameters
        parameters_opt = dyn.todict()
        parameters_opt.pop("logfile", None)
        parameters_opt.pop("restart", None)

        opt_fields = {
            "parameters_opt": parameters_opt,
            "converged": is_converged,
            "trajectory": atoms_trajectory,
            "trajectory_results": trajectory_results,
        }

        # Create a dictionary of the inputs/outputs
        unsorted_task_doc = base_task_doc | opt_fields | self.additional_fields

        return clean_dict(unsorted_task_doc)

    def md(
        self, dyn: MolecularDynamics, trajectory: list[Atoms] | None = None
    ) -> DynSchema:
        """
        Get tabulated results from an ASE MD run.

        Parameters
        ----------
        dyn
            ASE MolecularDynamics object.
        trajectory
            ASE Trajectory object or list[Atoms] from reading a trajectory file.

        Returns
        -------
        DynSchema
            Dictionary representation of the task document
        """
        # Check and set up variables
        base_task_doc = self.opt(dyn, trajectory=trajectory, check_convergence=False)
        del base_task_doc["converged"]

        # Clean up the opt parameters
        parameters_md = base_task_doc.pop("parameters_opt")
        parameters_md.pop("logfile", None)

        trajectory_log = []
        for t, atoms in enumerate(base_task_doc["trajectory"]):
            trajectory_log.append(
                {
                    "kinetic_energy": atoms.get_kinetic_energy(),
                    "temperature": atoms.get_temperature(),
                    "time": t * parameters_md["timestep"],
                }
            )

        md_fields = {"parameters_md": parameters_md, "trajectory_log": trajectory_log}

        # Create a dictionary of the inputs/outputs
        unsorted_task_doc = base_task_doc | md_fields | self.additional_fields

        return clean_dict(unsorted_task_doc)

    def neb(
        self,
        dyn: Optimizer,
        n_images: int,
        n_iter_return: int = -1,
        trajectory: TrajectoryWriter | list[Atoms] | None = None,
    ) -> OptSchema:
        """
        Summarize the NEB run results and store them in a database-friendly format.

        Parameters
        ----------
        dyn
            ASE Optimizer object used for the NEB run.
        n_images
            Number of images in the NEB run.
        n_iter_return
            Number of iterations to return. If -1, all iterations are returned.
        trajectory
            Trajectory of the NEB run, either as a Trajectory object or a list of Atoms objects.

        Returns
        -------
        OptSchema
            A dictionary containing the summarized NEB run results.
        """

        # Get trajectory
        atoms_trajectory = (
            trajectory or read(dyn.trajectory, index=":")
            if isinstance(dyn.trajectory, (str, Path))
            else read(dyn.trajectory.filename, index=":")
        )

        if n_iter_return == -1:
            atoms_trajectory = atoms_trajectory[-(n_images):]
        else:
            atoms_trajectory = _get_nth_iteration(
                atoms_trajectory,
                int(len(atoms_trajectory) / n_images),
                n_images,
                n_iter_return,
            )
        trajectory_results = [atoms.calc.results for atoms in atoms_trajectory]
        ts_index = (
            np.argmax(
                [
                    result["energy"]
                    for result in trajectory_results[-(n_images - 1) : -1]
                ]
            )
            + 1
        )
        ts_atoms = atoms_trajectory[ts_index]
        base_task_doc = atoms_to_metadata(atoms_trajectory[0])

        # Clean up the opt parameters
        parameters_opt = dyn.todict()
        parameters_opt.pop("logfile", None)
        parameters_opt.pop("restart", None)

        opt_fields = {
            "parameters_opt": parameters_opt,
            "trajectory": atoms_trajectory,
            "trajectory_results": trajectory_results,
            "ts_atoms": ts_atoms,
        }

        # Create a dictionary of the inputs/outputs
        unsorted_task_doc = base_task_doc | opt_fields | self.additional_fields

        return clean_dict(unsorted_task_doc)


class VibSummarize:
    """
    Summarize an ASE Vibrations analysis.
    """

    def __init__(
        self,
        vib_object: Vibrations | VibrationsData,
        directory: str | Path | None = None,
        additional_fields: dict[str, Any] | None = None,
    ) -> None:
        """
        Initialize the Summarize object.

        Parameters
        ----------
        vib_object
            Instantiated ASE Vibrations object.
        directory
            Path to the directory where the results will be stored.
        additional_fields
            Additional fields to add to the task document.

        Returns
        -------
        None
        """
        self.vib_object = vib_object
        self.directory = directory or "."
        self.additional_fields = additional_fields or {}
        self._settings = get_settings()

    def vib(self, *, is_molecule: bool) -> VibSchema:
        """
        Get tabulated results from an ASE Vibrations object and store them in a database-
        friendly format.

        Parameters
        ----------
        is_molecule
            Whether the Atoms object is a molecule. If True, the vibrational modes are
            sorted by their absolute value and the 3N-5 or 3N-6 modes are taken. If False,
            all vibrational modes are taken.

        Returns
        -------
        VibSchema
            Dictionary representation of the task document
        """
        # Tabulate input parameters
        vib_freqs_raw = self.vib_object.get_frequencies().tolist()
        vib_energies_raw = self.vib_object.get_energies().tolist()
        if isinstance(self.vib_object, VibrationsData):
            atoms = self.vib_object._atoms
            directory = self.directory
            inputs = {"nid": get_uri(directory).split(":")[0], "dir_name": directory}
        else:
            atoms = self.vib_object.atoms
            directory = self.directory or atoms.calc.directory
            inputs = {
                "parameters": atoms.calc.parameters,
                "parameters_vib": {
                    "delta": self.vib_object.delta,
                    "direction": self.vib_object.direction,
                    "method": self.vib_object.method,
                    "ndof": self.vib_object.ndof,
                    "nfree": self.vib_object.nfree,
                },
                "nid": get_uri(directory).split(":")[0],
                "dir_name": directory,
                "quacc_version": __version__,
            }

        # Convert imaginary modes to negative values for DB storage
        for i, f in enumerate(vib_freqs_raw):
            if np.imag(f) > 0:
                vib_freqs_raw[i] = -np.abs(f)
                vib_energies_raw[i] = -np.abs(vib_energies_raw[i])
            else:
                vib_freqs_raw[i] = np.abs(f)
                vib_energies_raw[i] = np.abs(vib_energies_raw[i])

        # Get the true vibrational modes
        atoms_metadata = atoms_to_metadata(atoms)

        natoms = len(atoms)
        if natoms == 1:
            vib_freqs = []
            vib_energies = []
        elif is_molecule:
            is_linear = (
                PointGroupData()
                .from_molecule(AseAtomsAdaptor().get_molecule(atoms))
                .linear
                if atoms.pbc.any()
                else atoms_metadata["molecule_metadata"]["symmetry"]["linear"]
            )

            # Sort by absolute value
            vib_freqs_raw_sorted = vib_freqs_raw.copy()
            vib_energies_raw_sorted = vib_energies_raw.copy()
            vib_freqs_raw_sorted.sort(key=np.abs)
            vib_energies_raw_sorted.sort(key=np.abs)

            # Cut the 3N-5 or 3N-6 modes based on their absolute value
            n_modes = 3 * natoms - 5 if is_linear else 3 * natoms - 6
            vib_freqs = vib_freqs_raw_sorted[-n_modes:]
            vib_energies = vib_energies_raw_sorted[-n_modes:]
        else:
            vib_freqs = vib_freqs_raw
            vib_energies = vib_energies_raw

        imag_vib_freqs = [f for f in vib_freqs if f < 0]

        vib_results = {
            "results": {
                "imag_vib_freqs": imag_vib_freqs,
                "n_imag": len(imag_vib_freqs),
                "vib_energies": vib_energies,
                "vib_freqs": vib_freqs,
                "vib_energies_raw": vib_energies_raw,
                "vib_freqs_raw": vib_freqs_raw,
            }
        }
        unsorted_task_doc = (
            atoms_metadata | inputs | vib_results | self.additional_fields
        )

        return clean_dict(unsorted_task_doc)

    def vib_and_thermo(
        self,
        thermo_method: Literal["ideal_gas", "harmonic"],
        energy: float = 0.0,
        temperature: float = 298.15,
        pressure: float = 1.0,
    ) -> VibThermoSchema:
        """
        Get tabulated results from an ASE Vibrations object and thermochemistry.

        Parameters
        ----------
        thermo_method
            Method to use for thermochemistry calculations. If None, no thermochemistry
            calculations are performed.
        energy
            Potential energy in eV used as the reference point for thermochemistry calculations.
        temperature
            Temperature in K for thermochemistry calculations.
        pressure
            Pressure in atm for thermochemistry calculations

        Returns
        -------
        VibThermoSchema
            Dictionary representation of the task document
        """
        atoms = (
            self.vib_object._atoms
            if isinstance(self.vib_object, VibrationsData)
            else self.vib_object.atoms
        )
        is_molecule = bool(thermo_method == "ideal_gas")

        # Generate vib data
        vib_schema = self.vib(is_molecule=is_molecule)

        # Generate thermo data
        thermo_summary = ThermoSummarize(
            atoms,
            vib_schema["results"]["vib_freqs_raw"],
            energy=energy,
            additional_fields=self.additional_fields,
        )
        if thermo_method == "ideal_gas":
            thermo_schema = thermo_summary.ideal_gas(
                temperature=temperature, pressure=pressure
            )
        elif thermo_method == "harmonic":
            thermo_schema = thermo_summary.harmonic(
                temperature=temperature, pressure=pressure
            )
        else:
            raise ValueError(f"Unsupported thermo_method: {thermo_method}.")

        # Merge the vib and thermo data
        unsorted_task_doc = recursive_dict_merge(vib_schema, thermo_schema)

        return clean_dict(unsorted_task_doc)


def _get_nth_iteration(
    neb_trajectory: list[Atoms], n_iter: int, n_images: int, interval: int
) -> list[Atoms]:
    """
    Extract every nth iteration from the NEB trajectory.

    Parameters
    ----------
    neb_trajectory
        List of configurations (length: n_iter * n_images).
    n_iter
        Total number of iterations.
    n_images
        Number of images per iteration.
    interval
        Interval to get every nth iteration.

    Returns
    -------
    list[Atoms]
        List of configurations from every nth iteration.
    """
    result = []
    start_idx, end_idx = 0, 0
    for i in range(0, n_iter, interval):
        start_idx = i * n_images
        end_idx = start_idx + n_images

        result.extend(neb_trajectory[start_idx:end_idx])
    if end_idx < len(neb_trajectory) - 1:
        result.extend(neb_trajectory[-(n_images):])
    return result



"""Core recipes for TorchSim."""

from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, Literal

from monty.dev import requires

from quacc import __version__, job
from quacc.recipes.torchsim._base import (
    pick_model,
    process_binning_autobatcher_dict,
    process_in_flight_autobatcher_dict,
    process_trajectory_reporter_dict,
)
from quacc.runners._base import BaseRunner
from quacc.schemas.torchsim import CONVERGENCE_FN_REGISTRY, TSModelType

has_torchsim = bool(find_spec("torch_sim"))
if has_torchsim:
    import torch_sim as ts

if TYPE_CHECKING:
    from pathlib import Path

    from ase.atoms import Atoms

    if has_torchsim:
        from torch_sim.integrators import Integrator
        from torch_sim.optimizers import Optimizer

    from quacc.recipes.torchsim._base import (
        AutobatcherDict,
        TorchSimIntegrateSchema,
        TorchSimOptSchema,
        TorchSimStaticSchema,
        TrajectoryReporterDict,
    )


@job
@requires(has_torchsim, "torch_sim is required for this function")
def relax_job(
    atoms: list[Atoms],
    model_type: TSModelType,
    model_path: str | Path,
    optimizer: Optimizer,
    *,
    convergence_fn: Literal["energy", "force"] = "force",
    trajectory_reporter_dict: TrajectoryReporterDict | None = None,
    autobatcher_dict: AutobatcherDict | bool = False,
    max_steps: int = 10_000,
    steps_between_swaps: int = 5,
    init_kwargs: dict[str, Any] | None = None,
    model_kwargs: dict[str, Any] | None = None,
    convergence_fn_kwargs: dict[str, Any] | None = None,
    **optimizer_kwargs: Any,
) -> TorchSimOptSchema:
    """
    Carry out a geometry optimization on a set of atoms.

    Parameters
    ----------
    atoms : list[Atoms]
        The list of atoms objects.
    model_type : TSModelType
        The type of model to use, limited to the types supported by TorchSim.
    model_path : str | Path
        The path to the model file or checkpoint.
    optimizer : Optimizer
        The TorchSim optimizer to use.
    convergence_fn : Literal["energy", "force"]
        The convergence function, either "energy" or "force". This will use either the
        ts.generate_energy_convergence_fn or ts.generate_force_convergence_fn function
        to interally generate the convergence function. Arguments can be supplied via
        the convergence_fn_kwargs argument. Used to select convergence function
        generators from quacc.schemas.torchsim.CONVERGENCE_FN_REGISTRY.
    trajectory_reporter_dict : TrajectoryReporterDict | None
        This dictionary defines the trajectory reporting behavior. This is a
        quacc-specific dictionary that allows for the configuration of the
        TrajectoryReporter. For a list of available keys, refer to the TorchSim
        TrajectoryReporter documentation or the TrajectoryReporterDict type-hint.
    autobatcher_dict : AutobatcherDict | bool
        This dictionary defines the autobatcher behavior. This is a quacc-specific
        dictionary that allows for the configuration of the autobatcher. For a list of
        available keys, refer to the TorchSim Autobatcher's documentation or the
        AutobatcherDict type-hint.
    max_steps : int
        The maximum number of steps to run for each optimization.
    steps_between_swaps : int
        Number of steps to take before checking convergence and swapping out states.
    init_kwargs : dict[str, Any] | None
        Keyword arguments passed to the optimizer initialization function.
    model_kwargs : dict[str, Any] | None
        Keyword arguments passed to the model.
    convergence_fn_kwargs : dict[str, Any] | None
        Keyword arguments passed to the convergence function generator.
    **optimizer_kwargs : Any
        Keyword arguments to pass to the optimizer step function.

    Returns
    -------
    TorchSimOptSchema
        A dictionary representing the final atoms configuration and metadata geometry
        optimization schema. See the type-hint for the data structure.
    """
    runner = BaseRunner()
    runner.setup()

    model = pick_model(model_type, model_path, **model_kwargs or {})

    trajectory_reporter, trajectory_reporter_details = process_trajectory_reporter_dict(
        trajectory_reporter_dict, runner, n_systems=len(atoms)
    )

    max_iterations = max_steps // steps_between_swaps
    autobatcher, autobatcher_details = process_in_flight_autobatcher_dict(
        atoms, model, autobatcher_dict=autobatcher_dict, max_iterations=max_iterations
    )

    convergence_fn_obj = CONVERGENCE_FN_REGISTRY[convergence_fn](
        **convergence_fn_kwargs or {}
    )

    state = ts.optimize(
        system=atoms,
        model=model,
        optimizer=optimizer,
        convergence_fn=convergence_fn_obj,
        trajectory_reporter=trajectory_reporter,
        autobatcher=autobatcher,
        max_steps=max_steps,
        steps_between_swaps=steps_between_swaps,
        init_kwargs=init_kwargs,
        **optimizer_kwargs,
    )
    runner.cleanup()

    return {
        "atoms": state.to_atoms(),
        "dir_name": str(runner.job_results_dir),
        "model_type": model_type,
        "model_path": model_path,
        "optimizer": optimizer,
        "convergence_fn": convergence_fn,
        "trajectory_reporter": trajectory_reporter_details,
        "autobatcher": autobatcher_details,
        "max_steps": max_steps,
        "steps_between_swaps": steps_between_swaps,
        "init_kwargs": init_kwargs,
        "model_kwargs": model_kwargs,
        "convergence_fn_kwargs": convergence_fn_kwargs,
        "optimizer_kwargs": optimizer_kwargs,
        "quacc_version": __version__,
    }


@job
@requires(has_torchsim, "torch_sim is required for this function")
def md_job(
    atoms: list[Atoms],
    model_type: TSModelType,
    model_path: str | Path,
    integrator: Integrator,
    *,
    n_steps: int,
    temperature: float | list,
    timestep: float,
    trajectory_reporter_dict: TrajectoryReporterDict | None = None,
    autobatcher_dict: AutobatcherDict | bool = False,
    model_kwargs: dict[str, Any] | None = None,
    **integrator_kwargs: Any,
) -> TorchSimIntegrateSchema:
    """
    Carry out a molecular dynamics calculation on a set of atoms.

    Parameters
    ----------
    atoms : list[Atoms]
        The list of atoms objects.
    model_type : TSModelType
        The type of model to use, limited to the types supported by TorchSim.
    model_path : str | Path
        The path to the model file or checkpoint.
    integrator : Integrator
        The TorchSim integrator to use.
    n_steps : int
        The maximum number of steps to run for each integration.
    temperature : float | list
        The temperature to use.
    timestep : float
        The timestep to use.
    trajectory_reporter_dict : TrajectoryReporterDict | None
        This dictionary defines the trajectory reporting behavior. This is a
        quacc-specific dictionary that allows for the configuration of the
        TrajectoryReporter. For a list of available keys, refer to the TorchSim
        TrajectoryReporter documentation or the TrajectoryReporterDict type-hint.
    autobatcher_dict : AutobatcherDict | bool
        This dictionary defines the autobatcher behavior. This is a quacc-specific
        dictionary that allows for the configuration of the autobatcher. For a list of
        available keys, refer to the TorchSim Autobatcher's documentation or the
        AutobatcherDict type-hint.
    model_kwargs : dict[str, Any] | None
        Keyword arguments passed to the model.
    **integrator_kwargs : Any
        Keyword arguments to pass to the integrator step function.

    Returns
    -------
    TorchSimIntegrateSchema
        A dictionary representing the final atoms configuration and metadata for the
        molecular dynamics schema. See the type-hint for the data structure.
    """
    runner = BaseRunner()
    runner.setup()

    model = pick_model(model_type, model_path, **model_kwargs or {})

    # Configure trajectory reporter
    trajectory_reporter, trajectory_reporter_details = process_trajectory_reporter_dict(
        trajectory_reporter_dict, runner, n_systems=len(atoms)
    )

    # Configure autobatcher
    autobatcher, autobatcher_details = process_binning_autobatcher_dict(
        atoms, model, autobatcher_dict=autobatcher_dict
    )

    state = ts.integrate(
        system=atoms,
        model=model,
        integrator=integrator,
        n_steps=n_steps,
        temperature=temperature,
        timestep=timestep,
        trajectory_reporter=trajectory_reporter,
        autobatcher=autobatcher,
        **integrator_kwargs,
    )
    runner.cleanup()

    return {
        "atoms": state.to_atoms(),
        "dir_name": str(runner.job_results_dir),
        "model_type": model_type,
        "model_path": model_path,
        "integrator": integrator,
        "n_steps": n_steps,
        "temperature": temperature,
        "timestep": timestep,
        "trajectory_reporter": trajectory_reporter_details,
        "autobatcher": autobatcher_details,
        "model_kwargs": model_kwargs,
        "integrator_kwargs": integrator_kwargs,
        "quacc_version": __version__,
    }


@job
@requires(has_torchsim, "torch_sim is required for this function")
def static_job(
    atoms: list[Atoms],
    model_type: TSModelType,
    model_path: str | Path,
    *,
    trajectory_reporter_dict: TrajectoryReporterDict | None = None,
    autobatcher_dict: AutobatcherDict | bool = False,
    model_kwargs: dict[str, Any] | None = None,
) -> TorchSimStaticSchema:
    """
    Carry out a static calculation on a set of atoms.

    Parameters
    ----------
    atoms : list[Atoms]
        The list of atoms objects.
    model_type : TSModelType
        The type of model to use, limited to the types supported by TorchSim.
    model_path : str | Path
        The path to the model file or checkpoint.
    trajectory_reporter_dict : TrajectoryReporterDict | None
        This dictionary defines the trajectory reporting behavior. This is a
        quacc-specific dictionary that allows for the configuration of the
        TrajectoryReporter. For a list of available keys, refer to the TorchSim
        TrajectoryReporter documentation or the TrajectoryReporterDict type-hint.
    autobatcher_dict : AutobatcherDict | bool
        This dictionary defines the autobatcher behavior. This is a quacc-specific
        dictionary that allows for the configuration of the autobatcher. For a list of
        available keys, refer to the TorchSim Autobatcher's documentation or the
        AutobatcherDict type-hint.
    model_kwargs : dict[str, Any] | None
        Keyword arguments passed to the model.

    Returns
    -------
    TorchSimStaticSchema
        A dictionary representing the final atoms configuration and metadata for the
        static calculation schema. See the type-hint for the data structure.
    """
    runner = BaseRunner()
    runner.setup()

    model = pick_model(model_type, model_path, **model_kwargs or {})

    trajectory_reporter, trajectory_reporter_details = process_trajectory_reporter_dict(
        trajectory_reporter_dict, runner, n_systems=len(atoms)
    )

    autobatcher, autobatcher_details = process_binning_autobatcher_dict(
        atoms, model, autobatcher_dict=autobatcher_dict
    )

    all_properties = ts.static(
        system=atoms,
        model=model,
        trajectory_reporter=trajectory_reporter,
        autobatcher=autobatcher,
    )

    all_properties_numpy = [
        {name: t.cpu().numpy() for name, t in prop_dict.items()}
        for prop_dict in all_properties
    ]
    runner.cleanup()

    return {
        "atoms": atoms,
        "dir_name": str(runner.job_results_dir),
        "all_properties": all_properties_numpy,
        "model_type": model_type,
        "model_path": model_path,
        "trajectory_reporter": trajectory_reporter_details,
        "autobatcher": autobatcher_details,
        "model_kwargs": model_kwargs,
        "quacc_version": __version__,
    }
